# This Powershell script can be run on archive server.
# Workflow installation is not required to run this script.
# This enables Workflow history option on SharePoint lists and libraries to view the Winshuttle workflow history.
#
# How to run ?
# Go to the Microsoft Sharepoint 2010/2013 Management Shell, run it as Administrator.
# Now browse to the directory where this script is placed. 
# Please make sure that WFHistoryViewer2013.wsp/WFHistoryViewer2010.wsp file is present in same location of script.
# Now type following command:
# 
#        .\Deploy_History_Viewer.ps1 
#
# Now press Enter. You are done !!!

$ver = (Get-PSSnapin microsoft.sharepoint.powershell).Version.Major

If ($ver -eq "15" ) 
    { 
        $wspname = 'WFHistoryViewer2013.wsp' 
    } 
 elseif ($ver -eq "14") 
    { 
        $wspname = 'WFHistoryViewer2010.wsp' 
	}
else
	{
		Write-Output "SharePoint server not found." 
	}	

Write-output $wspname;

$wsp = Get-SPSolution $wspname -ErrorAction silentlycontinue

Write-Output $wsp
if ($wsp -ne $null)
{
	if((Get-SPSolution $wsp).Deployed)
	{
		Uninstall-SPSolution $wsp -Confirm:$false 
		
		Write-Output "Uninstalling $wsp"
		do {start-sleep 2} while ($wsp.JobExists)
	}
	
	Remove-SPSolution $wsp -Confirm:$false
	Write-Output "$wsp removed." 
}

$path= Resolve-Path .\$wspname;

$wsp=Add-SPSolution -LiteralPath $path;
Write-Output "$wsp added."

If ($ver -eq "15" )
{
	Install-SPSolution $wsp  -CompatibilityLevel {14,15} -GACDeployment -Force
}
else
{
	Install-SPSolution $wsp  -GACDeployment -Force       
}

Write-Output "Installing $wsp."
do {start-sleep 2} while (((Get-SPSolution $wsp).JobExists))

Write-Output "Completed successfully."


 